"use strict";
var Promise = require("bluebird");
var utils = require("./utils");
var WinReg = require("winreg");
var which = require("which");
var async_1 = require("async");
var path_1 = require("path");
var underscore_1 = require("underscore");
var child_process_1 = require("child_process");
var util_1 = require("util");
var JavaVersion = (function () {
    function JavaVersion(version) {
        var matches = version.match(/(\d+?)\.(\d+?)\.(\d+?)(?:_(\d+))?/);
        this.major = parseInt(matches[1]);
        this.minor = parseInt(matches[2]);
        this.patch = parseInt(matches[3]);
        this.update = parseInt(matches[4] || "0");
    }
    return JavaVersion;
}());
exports.JavaVersion = JavaVersion;
var JavaInstall = (function () {
    function JavaInstall(path) {
        this._path = path;
        this._gotInfo = false;
        this._invalid = false;
    }
    Object.defineProperty(JavaInstall.prototype, "path", {
        get: function () {
            return this._path;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(JavaInstall.prototype, "arch", {
        get: function () {
            return this._arch;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(JavaInstall.prototype, "version", {
        get: function () {
            return this._version;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(JavaInstall.prototype, "invalid", {
        get: function () {
            return this._invalid;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @internal
     */
    JavaInstall.prototype.ensureInfo = function () {
        var _this = this;
        if (!this._gotInfo && !this._invalid) {
            return new Promise(function (resolve, reject) {
                var jarFile = __dirname + "/java/PrintJavaVersion.jar";
                debug("jarFile: " + jarFile);
                child_process_1.execFile(_this._path, ["-jar", jarFile], { timeout: 1000 }, function (err, stdoutBuf, stderrBuf) {
                    if (err) {
                        debug("[" + _this._path + "] Err: " + err);
                        _this._invalid = true;
                        resolve();
                        return;
                    }
                    var stdout = stdoutBuf.toString().trim();
                    debug("[" + _this._path + "] stdout: " + util_1.inspect(stdout));
                    var lines = stdout.split("\n");
                    var arch = lines[1];
                    switch (arch) {
                        case "32":
                            _this._arch = "x86";
                            break;
                        case "64":
                            _this._arch = "x64";
                            break;
                        default:
                            _this._arch = "unknown";
                            break;
                    }
                    var version = lines[0];
                    _this._version = new JavaVersion(version);
                    debug("[" + _this._path + "] " + util_1.inspect(_this));
                    _this._gotInfo = true;
                    resolve();
                });
            });
        }
        else {
            return Promise.resolve();
        }
    };
    return JavaInstall;
}());
exports.JavaInstall = JavaInstall;
var debug = function (debug) { };
function setDebug(debugFn) {
    debug = debugFn;
}
exports.setDebug = setDebug;
exports.getJavas = utils.PromiseCache(function () {
    debug("getJavas start (" + process.platform + ")");
    var javas;
    switch (process.platform) {
        case "win32":
            javas = findJavasWindows();
            break;
        case "darwin":
            javas = findJavasMac();
            break;
        case "linux":
            javas = findJavasLinux();
            break;
        default:
            javas = findJavaOnPath();
            break;
    }
    return javas
        .tap(function (v) { debug("Versions Raw: " + util_1.inspect(v.map(function (v) { return v.path; }))); })
        .filter(function (version) { return utils.canExecute(version.path); })
        .tap(function (v) { debug("Versions Existing: " + util_1.inspect(v.map(function (v) { return v.path; }))); })
        .then(function (versions) { return underscore_1.unique(versions, function (v) { return v.path; }); })
        .each(function (version) { return version.ensureInfo(); })
        .filter(function (version) { return !version.invalid; })
        .tap(function (v) { debug("Versions Final: " + util_1.inspect(v)); });
});
//region Linux
var defaultJavasLinux = [
    new JavaInstall("/opt/java/bin/java"),
    new JavaInstall("/usr/bin/java")
];
function findJavasLinux() {
    return Promise.all([defaultJavasLinux, findJavaOnPath()]).then(utils.flatten);
}
//endregion
//region Mac
var defaultJavasMac = [
    new JavaInstall("/Applications/Xcode.app/Contents/Applications/Application Loader.app/Contents/MacOS/itms/java/bin/java"),
    new JavaInstall("/Library/Internet Plug-Ins/JavaAppletPlugin.plugin/Contents/Home/bin/java"),
    new JavaInstall("/System/Library/Frameworks/JavaVM.framework/Versions/Current/Commands/java")
];
function findJavasMac() {
    var javaVersionPromises = [];
    javaVersionPromises.push(findJavaOnPath());
    javaVersionPromises.push(Promise.resolve(defaultJavasMac));
    javaVersionPromises.push(utils.allDirectories("/Library/Java/JavaVirtualMachines/").catch(function (err) { return []; })
        .map(function (dir) { return [
        new JavaInstall(path_1.join(dir, "Contents/Home/bin/java")),
        new JavaInstall(path_1.join(dir, "Contents/Home/jre/bin/java"))
    ]; }).then(utils.flatten));
    javaVersionPromises.push(utils.allDirectories("/System/Library/Java/JavaVirtualMachines/").catch(function (err) { return []; })
        .map(function (dir) { return [
        new JavaInstall(path_1.join(dir, "Contents/Home/bin/java")),
        new JavaInstall(path_1.join(dir, "Contents/Commands/java"))
    ]; }).then(utils.flatten));
    return Promise.all(javaVersionPromises).then(utils.flatten);
}
//endregion
//region Windows
var javaRegKeys = [
    "SOFTWARE\\JavaSoft\\Java Runtime Environment",
    "SOFTWARE\\JavaSoft\\Java Development Kit"
];
var defaultJavasWindows = [
    new JavaInstall("C:/Program Files/Java/jre8/bin/javaw.exe"),
    new JavaInstall("C:/Program Files/Java/jre7/bin/javaw.exe"),
    new JavaInstall("C:/Program Files/Java/jre6/bin/javaw.exe"),
    new JavaInstall("C:/Program Files (x86)/Java/jre8/bin/javaw.exe"),
    new JavaInstall("C:/Program Files (x86)/Java/jre7/bin/javaw.exe"),
    new JavaInstall("C:/Program Files (x86)/Java/jre6/bin/javaw.exe")
];
function findJavasWindows() {
    var javaVersionPromises = [];
    javaVersionPromises.push(Promise.resolve(defaultJavasWindows));
    javaVersionPromises.push(findJavaOnPath());
    javaRegKeys.forEach(function (key) {
        javaVersionPromises.push(findJavasFromRegistryKey(key, "x64").catch(function (err) { return []; }));
        javaVersionPromises.push(findJavasFromRegistryKey(key, "x86").catch(function (err) { return []; }));
    });
    return Promise.all(javaVersionPromises).then(utils.flatten);
}
function findJavasFromRegistryKey(keyName, arch) {
    return new Promise(function (resolve, reject) {
        var key = new WinReg({ key: keyName, arch: arch });
        // For each subkey of the given key, each of which should be
        key.keys(function (err, javaKeys) {
            if (err) {
                resolve([]);
                return;
            }
            var javaVersions = [];
            async_1.each(javaKeys, function (javaKey, cb) {
                javaKey.get("JavaHome", function (err, javaHome) {
                    if (err)
                        return;
                    var path = path_1.join(javaHome.value, "bin", "javaw.exe");
                    javaVersions.push(new JavaInstall(path));
                    cb();
                });
            }, function (err) {
                debug("Reg key " + keyName + " arch " + arch + ", got " + javaVersions);
                if (err)
                    reject(err);
                else
                    resolve(javaVersions);
            });
        });
    });
}
//endregion
var whichP = Promise.promisify(which);
function findJavaOnPath() {
    return whichP("java")
        .then(function (path) { return [new JavaInstall(path)]; }, function (err) { return []; });
}
